library("biomaRt")
library("GenomicRanges")
library("zoo")

source("./src/FLOSS_score/footprint-assignment.R")
source("./src/FLOSS_score/floss.R")
source("./src/FLOSS_score/mouse-annotation.R")
source("./src/FLOSS_score/plot-utils.R")

Floss <- read.csv("results/chx-floss.txt", check.names=FALSE, row.names=1)
row.names(Floss) <- gsub("\\.[[:digit:]]*", "", row.names(Floss))

xref <- match(row.names(Floss), annots$ensembl_transcript_id)
table(is.na(xref))
trxs <- cbind(Floss, annots[xref,])
trxs$name <- row.names(trxs)

pcoding   = trxs[trxs$name %in% codingIds & !(trxs$name %in% mitoIds) & !(trxs$name %in% dirtyPCodingIds),]
mitoch    = trxs[trxs$name %in% codingIds &  (trxs$name %in% mitoIds),]
lincs     = trxs[trxs$name %in% lincIds ,]
snolincs  = trxs[trxs$name %in% snohgIds  &   trxs$transcript_biotype == "lincRNA",]
miscs     = trxs[trxs$name %in% miscIds,]

Cutoffs <- flossCutoffs(pcoding$cdsNRead, pcoding$cdsScore)
predx <- seq(1,6,0.1)

plotBase <- function(title) {
  plot(log10(pcoding$cdsNRead), pcoding$cdsScore, xlim=c(0,6), ylim=c(0,1),
       pch=".", col=solarGrey01,
       axes=F, xlab="Total Reads", ylab="Fragment Length Distribution Score", cex.lab=1.33, main = title)
  axis(1, log10(10**seq(0,6,1)), labels=c("1", "10", "100", "1k", "10k", "100k", "1M"), cex.axis=1)
  axis(1, log10(c(sapply(10**seq(0,5,1), minor_decade))), labels=F, tcl=-0.5)
  axis(2, seq(0,1,0.2))
}

plotDrug <- function() {
  plotBase("Mito CDS, mics and lincs")
  points(log10(mitoch$cdsNRead), mitoch$cdsScore,
         pch=21, cex=0.75, col=solarCyan, bg=lighten(solarCyan))
  points(log10(miscs$trxNRead), miscs$trxScore,
         pch=21, cex=0.75, col=solarYellow, bg=lighten(solarYellow))
  text(log10(miscs$trxNRead), miscs$trxScore,
       labels=miscs$external_gene_id, col=solarYellow, pos=4 )
  points(log10(lincs$trxNRead), lincs$trxScore,
         pch=21, cex=0.75, col=solarMagenta, bg=lighten(solarMagenta))
  
  plotBase("5'UTR of protein coding mRNA")
  points(log10(pcoding$utr5NRead), pcoding$utr5Score
         , pch=21, cex=0.25, col=solarRed, bg=lighten(solarRed))
}

pdf("results/10_FLOSS_score/hg38_NY5-floss-rplot.pdf", useDingbats=F)
plotDrug()

plotBase("5'UTR")
lines(predx, predict(Cutoffs$extremeLoess, predx))

pcoding$classifyUtr5 <- Cutoffs$classify(pcoding$utr5NRead, pcoding$utr5Score)
utr5Good <- pcoding[pcoding$classifyUtr5 != "Extreme" & !is.na(pcoding$classifyUtr5),]
utr5Bad <- pcoding[pcoding$classifyUtr5 == "Extreme" & !is.na(pcoding$classifyUtr5),]

points(log10(utr5Good$utr5NRead), utr5Good$utr5Score
       , pch=21, cex=0.25, col=solarRed, bg=lighten(solarRed))
points(log10(utr5Bad$utr5NRead), utr5Bad$utr5Score
       , pch=21, cex=0.25, col=solarCyan, bg=lighten(solarCyan))

plotBase("lincs")
lines(predx, predict(Cutoffs$extremeLoess, predx))

lincs$classify <- Cutoffs$classify(lincs$trxNRead, lincs$trxScore)
lincsGood <- lincs[lincs$classify != "Extreme" & !is.na(lincs$classify),]
lincsBad <- lincs[lincs$classify == "Extreme" & !is.na(lincs$classify),]

points(log10(lincsGood$trxNRead), lincsGood$trxScore,
       pch=21, cex=0.25, col=solarMagenta, bg=lighten(solarMagenta))
points(log10(lincsBad$trxNRead), lincsBad$trxScore,
       pch=21, cex=0.25, col=solarGreen, bg=lighten(solarGreen))

dev.off()

pcoding$classifyCds <- Cutoffs$classify(pcoding$cdsNRead, pcoding$cdsScore)
table(pcoding[pcoding$cdsNRead > 100, "classifyCds"])
sum(pcoding[pcoding$cdsNRead > 100, "classifyCds"] != "Extreme", na.rm=TRUE) / sum(!is.na(pcoding[pcoding$cdsNRead > 100, "classifyCds"]))

table(pcoding[pcoding$utr5NRead > 100, "classifyUtr5"])
sum(pcoding[pcoding$utr5NRead > 100, "classifyUtr5"] != "Extreme", na.rm=TRUE) / sum(!is.na(pcoding[pcoding$utr5NRead > 100, "classifyUtr5"]))

table(lincs[lincs$trxNRead > 100, "classify"])
sum(lincs[lincs$trxNRead > 100, "classify"] != "Extreme", na.rm=TRUE) / sum(!is.na(lincs[lincs$trxNRead > 100, "classify"]))

dir.create("results/tables", showWarnings = FALSE)

write.csv(pcoding[,c("cdsNRead", "cdsScore", "classifyCds")],
          "results/tables/floss-refcoding-cds.txt")
write.csv(pcoding[,c("utr5NRead", "utr5Score", "classifyUtr5")],
          "results/tables/floss-refcoding-utr5.txt")
write.csv(lincs[,c("trxNRead", "trxScore", "classify")],
          "results/tables/floss-lincs.txt")
write.csv(mitoch[,c("cdsNRead", "cdsScore")],
          "results/tables/floss-mitoch.txt")
write.csv(miscs[,c("trxNRead", "trxScore")],
          "results/tables/floss-miscs.txt")
write.csv(lincs[,c("ensembl_gene_id", "trxNRead", "trxScore", "classify", "external_gene_name", "gene_biotype", "transcript_biotype", "chromosome_name", "start_position", "end_position", "strand", "description")],
          "results/tables/floss-lincs-annotated.txt")
