library("zoo")

## Length range to consider: 26 through 34 nucleotides
defaultLengthMin <- 26
defaultLengthMax <- 34

## Length counts for a GAlignments
alignLengths <- function(galns, lengthMin=defaultLengthMin, lengthMax=defaultLengthMax) {
    lmax = 1 + lengthMax - lengthMin
    ld <- tabulate(pmax(0, pmin(lmax, qwidth(galns) - (lengthMin - 1))), nbins=lmax)
    names(ld) <- as.character(seq(lengthMin, lengthMax))
    ld
}

normLDist <- function(ld) { ld/sum(ld) }

## L1 norm length distribution difference between two length counts
ldistDiff <- function(d1, d2) { sum(abs(colSums(rbind(d1/sum(d1), - d2/sum(d2))))) / 2.0 }

trxRegionLengthDists <- function(bamfile, trxRgns, lengthMin=defaultLengthMin, lengthMax=defaultLengthMax) {
  alns <- getAlignments(bamfile, trxRgns$trx)

  regionLengthDist <- function(rgn) {
      alignLengths(alns[queryHits(findOverlaps(alns, rgn, type="within"))], lengthMin, lengthMax)
  }
  
  lapply( trxRgns, regionLengthDist )
}

calcTrxRegionLengthDists <- function(bamfile, trxBed, verbose=NA,
                                     lengthMin=defaultLengthMin, lengthMax=defaultLengthMax, ...) {
    ttl <- length(trxBed)
    calcOne <- function(i) {
        if (!is.na(verbose) && i %% verbose == 0) { cat(sprintf("%d\t%d\t%s\t%s\n", i, ttl, as.character(trxBed$name[i]), path(bamfile))) }
        trxGR <- transcriptGRanges(trxBed[i])
        cdsIR <- transcriptCdsIRanges(trxGR, start(trxBed$thick)[[i]], end(trxBed$thick)[[i]])
        trxRegionLengthDists(bamfile, trxRegions(trxGR, cdsIR, ...), lengthMin, lengthMax)
    }
    ld <- mclapply(seq(1,ttl), calcOne)
    names(ld) <- trxBed$name
    ld
}

sliceRegionLengthDists <- function(trxLDist, region) {
    df <- do.call(rbind.data.frame, lapply(trxLDist, function(r) { r[[region]] }))
    colnames(df) <- names(trxLDist[[1]][[region]])
    j=1
    for (i in 1:length(names(trxLDist))){
      if ( !is.null(trxLDist[[i]][[region]])) {
        rownames(df)[j] = names(trxLDist)[i]
        j=j+1
      }
     
    }
    df
}

writeRegionLengthDists <- function(trxLDist, pfx) {
    lapply(c("trx", "cds", "utr5", "utr3"),
           function(rgn) { rld <- sliceRegionLengthDists(trxLDist, rgn)
                           write.csv(rld, sprintf("%s-%s-ldist.txt", pfx, rgn), row.names=T)
                           sum(rld)
                       } )
}

readLengthDists <- function( pfx ) {
    read.csv(sprintf("%s-ldist.txt", pfx), row.names=1)
}

readRegionLengthDists <- function( pfx ) {
    list( trx = readLengthDists(sprintf("%s-trx", pfx)),
          cds = readLengthDists(sprintf("%s-cds", pfx)),
          utr5 = readLengthDists(sprintf("%s-utr5", pfx)),
          utr3 = readLengthDists(sprintf("%s-utr3", pfx)) )
}

flossLengthDist <- function( ldists, ldistRef ) {
    floss <- apply(ldists, 1, function(l) { ldistDiff(l, ldistRef) })
    nread <- apply(ldists, 1, function(l) { sum(l) })
    ldists$Score <- floss
    ldists$NRead <- nread
    ldists
}

flossRegionLengthDists <- function( rgnLDists, ldistRef) {
    list( trx = flossLengthDist( rgnLDists$trx, ldistRef ),
          cds = flossLengthDist( rgnLDists$cds, ldistRef ),
          utr5 = flossLengthDist( rgnLDists$utr5, ldistRef ),
          utr3 = flossLengthDist( rgnLDists$utr3, ldistRef ) )
}

flossRegionScores <- function( rgnFloss ) {
    floss <- data.frame( trxNRead = rgnFloss$trx$NRead,
                         trxScore = rgnFloss$trx$Score,
                         row.names = row.names(rgnFloss$trx) )
    floss[row.names(rgnFloss$cds),"cdsNRead"] <- rgnFloss$cds$NRead
    floss[row.names(rgnFloss$cds),"cdsScore"] <- rgnFloss$cds$Score
    floss[row.names(rgnFloss$utr5),"utr5NRead"] <- rgnFloss$utr5$NRead
    floss[row.names(rgnFloss$utr5),"utr5Score"] <- rgnFloss$utr5$Score
    floss[row.names(rgnFloss$utr3),"utr3NRead"] <- rgnFloss$utr3$NRead
    floss[row.names(rgnFloss$utr3),"utr3Score"] <- rgnFloss$utr3$Score
    floss
}

flossCutoffs <- function(nreads, score) {
  rollwidth <- 200

  unsorted <- data.frame(nreads=nreads, score=score)
  fr <- unsorted[order(unsorted$nreads),]
  nr <- rollapply(fr$nreads, width=rollwidth, function(xs) { mean(xs, na.rm=T) })
  medsc <- rollapply(fr$score, width=rollwidth, function(xs) { median(xs, na.rm=T) })
  q1 <- rollapply(fr$score, width=rollwidth, function(xs) { quantile(xs, probs=0.25, na.rm=T) })
  q3 <- rollapply(fr$score, width=rollwidth, function(xs) { quantile(xs, probs=0.75, na.rm=T) })
  rawExtreme <- q3 + 3.0*(q3-q1)
  extLoess <- loess(extreme ~ nread, data.frame(nread=log10(nr), extreme=rawExtreme))

  classify <- function(nr, sc) {
    extCutoff <- predict(extLoess, log10(nr))
    factor(
      if (is.na(extCutoff)) {
        NA
      } else if (sc > extCutoff) {
        "Extreme"      
      } else {
        "Good"
      },
      levels=c("Good", "Extreme")
      )
  }
  
  list( l10nreads = log10(nr),
       scoreMedian = medsc, scoreQ3 = q3, scoreExtreme = rawExtreme,
       extremeLoess = extLoess,
       isExtreme = function(nr, sc) { sc > predict(extLoess, log10(nr)) },
       classify = Vectorize(classify)
       )
}
