library("IRanges")
library("GenomicRanges")
library("Rsamtools")

# Convert an absolute genomic query position (qpos) to a
# transcript-relative coordinate position within a transcript whose
# coordinates are given by a GRanges of exons (outer).  NOTE that the
# exons in outer are in ascending numeric order regardless of the
# strand.
relativeWithin <- function(qpos, outer) {
  outerOffsets <- start(outer)
  withinOffsets <- head(c(0, cumsum(width(outer))), -1)

  startHits <- findOverlaps(IRanges(start=qpos, width=1), ranges(outer), select="first")
  
  within <- (qpos - outerOffsets[startHits]) + withinOffsets[startHits]

  if (as.vector(strand(outer))[1] == "-") {
    within <- (sum(width(outer)) - 1) - within
  }
  within
}

# Convert a transcript-relative query position (qpos) to an absolute
# genomic coordinate based on a transcript whose coordinates are given
# by a GRanges of exons (outer). NOTE that the exons are
# in ascending numeric order regardless of the strand.
absoluteOuter <- function(qpos, outer) {
  outerOffsets <- start(outer)
  withinOffsets <- head(c(0, cumsum(width(outer))), -1)
  inner <- IRanges(start=withinOffsets, width=width(outer))

  if (as.vector(strand(outer))[1] == "-") {
    qpos <- (sum(width(outer)) - 1) - qpos
  }

  startHits <- findOverlaps(IRanges(start=qpos, width=1), inner, select="first")

  (qpos - withinOffsets[startHits]) + outerOffsets[startHits]
}

# Convert transcript-relative query IRanges (qranges) to an absolute
# genomic GRanges based on a transcript whose coordinate are given by
# a GRanges of exons (outer).
irangeOuter <- function(qranges, outer) {
    starts <- absoluteOuter(start(qranges), outer)
    ends <- absoluteOuter(end(qranges), outer)

    if (as.vector(strand(outer))[1] == "-") {
        masks <- GRanges( seqnames = as.vector(seqnames(outer))[[1]],
                         ranges = IRanges( start = ends, end = starts ),
                         strand = "-" )
    } else {
        masks <- GRanges( seqnames = as.vector(seqnames(outer))[[1]],
                         ranges = IRanges( start = starts, end = ends ),
                         strand = "+" )        
    }

    intersect(masks, outer)
}

# relativeWithin(qpos) = relativeWithin(absoluteOuter(relativeWithin(qpos)))
# absoluteOuter(qpos) = absoluteOuter(relativeWithin(absoluteOuter(qpos)))
# relativeWithin(qpos) and absoluteOuter(qpos) may be NA.

# bedGRange should be a 1-element GRanges slice.
# The metadata columns should include a "blocks" entry as from an rtracklayer BED file.
# A GRanges object is returned with one range per exon.
# NOTE that the exons are in ascending numeric order regardless of the transcript strand.
transcriptGRanges <- function(bedGRange) {
    blocks <- mcols(bedGRange)$blocks[[1]]
    exons <- shift(blocks, shift = start(bedGRange)[[1]] - 1)
    sn <- seqnames(bedGRange)[1]
    runLength(sn) <- length(exons)
    st <- strand(bedGRange)[1]
    runLength(st) <- length(exons)
    
    GRanges(seqnames = sn,
            ranges = exons,
            strand = st)
}

# trx is a transcript GRanges wiht one range per exon, and thickStart
# & thickEnd are the genomic start and end of the coding sequence,
# e.g. taken from start(...) and end(...) on the $thick IRanges from a
# bed file.
# Return an IRanges for the CDS in transcript-local coordinates.
transcriptCdsIRanges <- function(trx, thickStart, thickEnd) {
    startWithin <- relativeWithin(thickStart, trx)
    endWithin <- relativeWithin(thickEnd, trx)

    if (is.na(startWithin) || is.na(endWithin)) {
        IRanges()
    } else {
        IRanges( start = min( startWithin, endWithin ), end = max( startWithin, endWithin ) )
    }
}

# For each transcript in a BED file and the associated transcript
# GRanges as computed by bedGRangesList, find the transcript-local CDS
# as per transcriptCdsIRanges and collect these as an IRangesList.
bedCdsIRangesList <- function(bed, bedgrl) {
    irl <- IRangesList(mclapply(1:length(bed), function(i) { transcriptCdsIRanges(bedgrl[[i]], start(bed$thick)[[i]], end(bed$thick)[[i]] ) } ))
    names(irl) <- bed$name
    irl
}

# For each transcript in a BED file GRanges, get the GRanges of its
# exons as per transcriptGRanges and collec these into a GRangesList
# with names taken from the name metadata column of the BED file
# GRanges.
bedGRangesList <- function(bed) {
    grl <- GRangesList(mclapply(1:length(bed), function(i) { transcriptGRanges(bed[i]) }))
    names(grl) <- bed$name
    grl
}

# Calculate a GRangesList with named GRanges giving genomic
# coordinates of transcript regions: trx for the entire transcript,
# cds for the coding sequence, utr5 for the 5' UTR, and utr3 for the
# 3' UTR.
trxRegions <- function(trx, cds,
                       insetUtr5Start=0, insetUtr5End=0,
                       insetCdsStart=0,  insetCdsEnd=0,
                       insetUtr3Start=0, insetUtr3End=0) {
    rgns <- list( trx = trx )

    if (length(cds) == 1) {
        utr5Start <- insetUtr5Start
        utr5End <- (start(cds) - 1) - insetUtr5End
        cdsStart <- start(cds) + insetCdsStart
        cdsEnd <- end(cds) - insetCdsEnd
        utr3Start <- (end(cds) + 1) + insetUtr3Start
        utr3End <- (sum(width(trx)) - 1) - insetUtr3End

        if (cdsStart < cdsEnd) {
            rgns$cds <- irangeOuter( IRanges( start = cdsStart, end = cdsEnd ), trx )
        }

        if (utr5Start < utr5End) {
            rgns$utr5 <- irangeOuter( IRanges( start = utr5Start, end = utr5End ), trx )
        }

        if (utr3Start < utr3End) {
            rgns$utr3 <- irangeOuter( IRanges( start = utr3Start, end = utr3End ), trx )
        }
    }

    rgns
    
}

# GAlignments containing all reads overlapping the genomic extent of a
#  primary transcript on the correct strand, including unspliced and
#  purely intronic reads.
getAllAlignments <- function(bamfile, trx) {
    range <- GRanges(seqnames = seqnames(trx)[1],
                     ranges = IRanges(start = min(start(trx)), end = max(end(trx))))
    
    isminus <- (as.vector(strand(trx))[1] == "-")
    
    readGAlignments(bamfile, param=ScanBamParam(which = range, flag=scanBamFlag(isMinusStrand = isminus)))
}

# GAlignments containing all reads compatible with a processed
#  transcript on the correct strand, as per findSpliceOverlaps.
getAlignments <- function(bamfile, trx) {
    alns <- getAllAlignments(bamfile, trx)
    
    spliceHits <- findSpliceOverlaps(alns, GRangesList(trx))
    compatibles <- spliceHits[mcols(spliceHits)$compatible]
    
    alns[queryHits(compatibles)]
}

# DNAString representing the transcript strand sequence of 
getSequence <- function(fafile, trx) {
    gseq <- unlist(scanFa(fafile, trx))
    if (as.vector(strand(trx))[1] == "-") {
        reverseComplement(gseq)
    } else {
        gseq
    }
}
